/* BYDATE.C - include the date in a file's name.

            Written by M.C.Gregorie.

            Version history
               
            1.0   10/08/07   Initial release
            2.0   13/12/08   Revised to use patterns
            3.0   04/06/09   Introduced the -t test option
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <environ.h>

#ifdef MSDOS
#include <unistd.h>
#include <dir.h>
#include <ctype.h>
#endif

#ifdef LINUX
#include <unistd.h>
#endif

#include "bydate.h"
#include "pattern.h"

#define PROGNAME  "bydate"
#define VERSION   "3.0"
#define COPYRIGHT "Martin Gregorie, (c) 2007,2008,2009"

/*
   Global variables
*/
extern int errno;

/*
   This function scans the command line for options and does
   all the associated setup.
*/
int readargs(int argc,
             char *argv[],
             int flags[],
             char **pattern,
             char **files)
{
   int         i;
   int         opt;
   static char *nullstr = "";

   /* initialise flags */

   for (i=0 ; i<FLAGCNT ; i++) flags[i] = FALSE;

   /* set up options from the command line */

   optmaybe = MAYBEMARK;
   optbad = BADMARK;
   optind = 1;
   *pattern = "cymd";
   *files = nullstr;
   
   while ((opt=getopt(argc,argv,OPTLIST)) != EOF)   
   {
      switch (opt) {
         case 0:        continue;
         case BADMARK:  return(opt);
         case '?':      flags[HELP] = TRUE;       
                        break;
         case 'd':      flags[DEBUG]++;
                        break;
         case 'p':      flags[PATTERN]++;
                        *pattern = (char*)get_space(strlen(optarg) + 1);
                        strcpy(*pattern,optarg);
                        break;
         case 't':      flags[TEST] = TRUE;
                        break;
         case 'z':      flags[FILENAMES]++;
                        if (optarg != NULL)
                        {
                           *files = (char*)get_space(strlen(optarg) + 1);
                           strcpy(*files,optarg);
                        }
                        break;
         default :      error(PROGNAME,
                              "impossible happened parsing options",
                              "");
      }
   }
   return(OK);
}

/*
   Show the punter what the options are
*/
void showhelp()
{
   printf("%s %s - %s\n\n", PROGNAME, VERSION, COPYRIGHT);
   printf("Syntax:   bydate [<opts>] filename... [<opts>]\n");
   printf("Function: Include the date in a file's name\n");
   printf("          by default file.ext becomes ccyymmdd.ext where\n");
   printf("          the string ccyymmdd is the current date.\n");
   printf("Options:\n");
   printf("          -d           Debug mode\n");
   printf("          -p=pattern   Set the pattern for the new filename.\n");
   printf("                       The default pattern is 'cymd'\n");
   printf("          -t           Test mode. Timestamp=2009-06-04 10:15:30\n");
   printf("          -z[<=path>]  read path names from stdin or <path> ");
   printf("if given\n");
}

/*
   The main program starts here
   ----------------------------
*/
int main(int argc, char *argv[])
{
   FILE  *pathlist;
   char  *pathsource,
         *inputfile,
         *pattern;
   PCB   *p = NULL;
   int   flags[FLAGCNT],
         pathresult,
         errors = 0,
         reply = 0;

   /*
      Parse and set up command options
   */
   setprog(PROGNAME);
   if (readargs(argc, argv, flags, &pattern, &pathsource) == BADMARK)
   {
      return(1);
   }

   /*
      If help was asked for, give it and quit
   */
   if (flags[HELP]) 
   {
      showhelp();
      return(0);
   }

   /*
      Validate the command options.
   */
   if (flags[PATTERN] > 1)
      error(PROGNAME, "only one pattern option is permitted", "");

   if (flags[FILENAMES] > 1)
      error(PROGNAME, "only one -z option is permitted", "");

   p = parse_pattern(pattern, flags[TEST], flags[DEBUG]);
   if (!pattern_ok(p))
      error(PROGNAME, "invalid filename pattern:", pattern_error(p));
   
   /*
      This code sets up for the -z option, which expects a list
      of the files to be processed on stdin or in a named file.
   */
   if (flags[FILENAMES])
   {
      if (strcmp(pathsource, "") == 0)
      {
         pathlist = fdopen(0, "r");
      }
      else
      {
         pathlist = fopen(pathsource,"r");
      }
      
      if (pathlist == NULL)
      {
         error(PROGNAME, "can't open path list", pathsource);
      }
   }

   /*
      The following code drives the application. 
   */
   optind = 1;
   while (inputfile = getpath(argc,argv,pathlist,flags[FILENAMES]))
   {
      if (processfile(inputfile, p, flags))
         errors++;
   }

   p = free_pattern(p);
   if (errors)
   {
      error(PROGNAME, "File(s) skipped", "");
      reply = 1;
   }
      
   return(reply);
}

/*
   processfile() - rename a single file
   =============
*/
int processfile(char *filename, PCB *p, int flags[])
{
   char        *inputname = 0;
   char        *iptr;
   char        *dirname = 0;
   char        *file = 0;
   char        *newname = 0;
   char        *thisdir = 0;
   int         errorcount = 0;

   /*
      Store the filename to be worked on
   */
   if (flags[DEBUG] > 1)
      fprintf(stderr, "processfile(%s,%s,flags)\n", filename, get_pattern(p));
 
   /*
      Make a working copy of the file name.
      Extract the directory name from the copy if it is present.
      Put the rest of the name in file.
   */
   inputname = dupstring(filename);
   iptr = strrchr(inputname, SEPCHAR);
   if (iptr)
   {
      /*
         Absolute or relative path found.
      */
      *iptr = 0;
      file = dupstring((iptr + 1));
      dirname = dupstring(inputname);
   }
   else
   {
      /*
         Only a simple filename.
      */
      file = dupstring(inputname);
      dirname = NULL;
   }
 
   /*
      Save the original file name then apply the pattern to
      generate the new file name.
   */
   newname = apply_pattern(p, file);
   if (!pattern_ok(p))
   {
      fprintf(stderr,
              "Skipping %s: error in %s: %s\n",
              filename,
              get_pattern(p),
              pattern_error(p));
      errorcount++;
   }
   else
   {
      if (flags[DEBUG] > 1)
         fprintf(stderr,
                 "Path:%s Old name:%s  New name:%s\n",
                 ( dirname ? dirname : "(none)"),
                 filename,
                 newname);

      thisdir = get_space(NAMELTH + 1);
      getcwd(thisdir, NAMELTH);
      if (flags[DEBUG] > 1)
         fprintf(stderr, "Current directory: %s\n", thisdir);

      changeto(dirname, flags[DEBUG]);
      if (access(newname, F_OK) == 0)
      {
         if (dirname)
            fprintf(stderr,
                    "Skipping %s: %s%c%s already exists\n",
                    filename,
                    dirname,
                    SEPCHAR,
                    newname);
         else
            fprintf(stderr,
                    "Skipping %s: %s already exists\n",
                    filename,
                    newname);
                    
         errorcount++;
      }
      else
      {
         if (flags[DEBUG])
         {
            if (dirname)
               fprintf(stderr,
                       "Renaming %s to %s in %s\n",
                       file,
                       newname,
                       dirname);
            else
               fprintf(stderr,
                       "Renaming %s to %s\n",
                       file,
                       newname);
         }
   
         if (rename(file, newname))
         {
            if (dirname)
               fprintf(stderr,
                       "Unable to rename %s to %s%c%s: %s\n",
                       filename,
                       dirname,
                       SEPCHAR,
                       newname,
                       strerror(errno));
            else
               fprintf(stderr,
                       "Unable to rename %s to %s: %s\n",
                       filename,
                       newname,
                       strerror(errno));
                       
            errorcount++;
         }
      }
      
      changeback(dirname, thisdir, flags[DEBUG]);
      thisdir = (char*)release_space(thisdir);
   }

   inputname = (char*)release_space(inputname);
   file = (char*)release_space(file);
   if (dirname)
      dirname = (char*)release_space(dirname);
      
   if (newname)
      newname = (char*)release_space(newname);

   if (flags[DEBUG] > 1)
      fprintf(stderr,
              "%d = processfile(%s,%s,flags)\n",
              errorcount,
              filename,
              get_pattern(p));
 
   return(errorcount);
}

/*
   Subordinate function declarations (in alphabetic order).
   ========================================================

   changeback() - change back to the original directory.
   ============   This is only needed by DOS/Windows programs.
                  Other implementations do nothing/
*/
void changeback(char *dir, char *original, int debug)
{
   int drive;

   if (debug > 2)
      fprintf(stderr, "changeback('%s','%s',%d)\n", dir, original, debug);
   
   if (dir)
   {
#ifdef MSDOS
      drive = toupper(original[0]) - 'A';
      setdisk(drive);
      if (debug > 1)
         fprintf(stderr, "Back in drive %c\n", drive + 'A');
#endif
      if (chdir(original))
         error(PROGNAME, "unable to change directory back to", original);

      if (debug > 1)
         fprintf(stderr,
                 "changeback('%s','%s',%d) switched back to %s\n",
                 dir,
                 original,
                 debug,
                 original);
   }
   else
   {
      if (debug > 1)
         fprintf(stderr,
                 "changeback(NULL,'%s',%d) did nothing\n",
                 dir,
                 original,
                 debug);
   }
}


/*
   changeto() - change to the directory containing the
   ==========   file to be renamed. 
*/
void changeto(char *dir, int debug)
{
   if (debug > 2)
      fprintf(stderr, "changeto('%s',%d)\n", dir, debug);

   if (!dir)
   {
      if (debug > 1)
         fprintf(stderr, "changeto(NULL,%d) did nothing\n", dir, debug);

      return;
   }
   
#ifdef MSDOS
   int thisdrive;
   int thatdrive;
   int drivecount;

   thisdrive = getdisk();
   thatdrive = toupper(dir[0]) - 'A';
   if (thisdrive != thatdrive)
      drivecount = setdisk(thatdrive);
   else
      drivecount = 0;

   if (debug > 1)
      fprintf(stderr,
              "Drives: current=%c, target=%c, count=%d\n",
              thisdrive + 'A',
              thatdrive + 'A',
              drivecount);
#endif
              
   if (chdir(dir))
      error(PROGNAME, "unable to change directory to", dir);

   if (debug > 1)
      fprintf(stderr,
              "changeto('%s',%d) switched reference to %s\n",
              dir,
              debug,
              dir);
}

/*
   dupstring() - make a copy of a string in heap storage.
   ===========
*/
char *dupstring(char *s)
{
   char *x;

   x = (char*)get_space(strlen(s) + 1);
   strcpy(x, s);
   return x;
}

